import pytest

from aiodocker.docker import Docker


@pytest.mark.asyncio
async def test_system_info(docker: Docker) -> None:
    docker_info = await docker.system.info()
    assert "ID" in docker_info
    assert "ServerVersion" in docker_info



from __future__ import annotations

import asyncio
import os
import sys
import traceback
import uuid
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
)

import pytest
from packaging.version import parse as parse_version
from typing_extensions import TypeAlias

from aiodocker.containers import DockerContainer
from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError


API_VERSIONS = {
    "17.06": "v1.30",
    "17.09": "v1.32",
    "17.12": "v1.35",
    "18.02": "v1.36",
    "18.03": "v1.37",
    "18.06": "v1.38",
    "18.09": "v1.39",
}


def _random_name():
    return "aiodocker-" + uuid.uuid4().hex[:7]


@pytest.fixture(scope="session")
async def random_name():
    yield _random_name

    # If some test cases have used randomly-named temporary images,
    # we need to clean up them!
    if os.environ.get("CI") is not None:
        # But inside the CI server, we don't need clean up!
        return

    docker = Docker()
    images = await docker.images.list()
    for img in images:
        if not img["RepoTags"]:
            continue
        try:
            if img["RepoTags"][0].startswith("aiodocker-"):
                print("Deleting image id: {}".format(img["Id"]))
                await docker.images.delete(img["Id"], force=True)
        except DockerError:
            traceback.print_exc()
    await docker.close()


@pytest.fixture(scope="session")
def image_name() -> str:
    if sys.platform == "win32":
        return "python:latest"
    return "python:alpine"


@pytest.fixture(scope="session")
async def testing_images(image_name: str) -> None:
    # Prepare a small Linux image shared by most test cases.
    docker = Docker()
    try:
        required_images = [image_name]
        for img in required_images:
            try:
                await docker.images.inspect(img)
            except DockerError as e:
                assert e.status == 404
                print(f'Pulling "{img}" for the testing session...')
                await docker.pull(img)
    finally:
        await docker.close()


@pytest.fixture
async def docker(testing_images):
    kwargs = {}
    version = os.environ.get("DOCKER_VERSION")
    if version:
        for k, v in API_VERSIONS.items():
            if version.startswith(k):
                kwargs["api_version"] = v
                break
        else:
            raise RuntimeError(f"Cannot find docker API version for {version}")

    docker = Docker(**kwargs)
    print(asyncio.get_running_loop())
    try:
        yield docker
    finally:
        await docker.close()


@pytest.fixture
async def requires_api_version(
    docker: Docker,
) -> AsyncIterator[Callable[[str, str], None]]:
    # Update version info from auto to the real value
    await docker.version()

    def check(version: str, reason: str) -> None:
        if parse_version(docker.api_version[1:]) < parse_version(version[1:]):
            pytest.skip(reason)

    yield check


@pytest.fixture
async def swarm(docker):
    if sys.platform == "win32":
        pytest.skip("swarm commands dont work on Windows")
    assert await docker.swarm.init()
    try:
        yield docker
    finally:
        assert await docker.swarm.leave(force=True)


AsyncContainerFactory: TypeAlias = Callable[
    [Dict[str, Any], str], Awaitable[DockerContainer]
]


@pytest.fixture
async def make_container(
    docker: Docker,
) -> AsyncIterator[AsyncContainerFactory]:
    container: DockerContainer | None = None

    async def _spawn(
        config: Dict[str, Any],
        name: str,
    ) -> DockerContainer:
        nonlocal container
        container = await docker.containers.create_or_replace(config=config, name=name)
        assert container is not None
        await container.start()
        return container

    try:
        yield _spawn
    finally:
        if container is not None:
            assert isinstance(container, DockerContainer)
            await container.delete(force=True)


@pytest.fixture
async def shell_container(
    docker: Docker,
    make_container,
    image_name: str,
) -> AsyncContainerFactory:
    config = {
        "Cmd": ["python"],
        "Image": image_name,
        "AttachStdin": False,
        "AttachStdout": False,
        "AttachStderr": False,
        "Tty": True,
        "OpenStdin": True,
    }

    return await make_container(config, name="aiodocker-testing-shell")



import sys

import pytest

from aiodocker.exceptions import DockerError


@pytest.mark.asyncio
async def test_swarm_inspect(swarm):
    swarm_info = await swarm.swarm.inspect()
    assert "ID" in swarm_info
    assert "Spec" in swarm_info


@pytest.mark.asyncio
async def test_swarm_failing_joining(swarm):
    swarm_info = await swarm.swarm.inspect()
    system_info = await swarm.system.info()
    swarm_addr = [system_info["Swarm"]["RemoteManagers"][-1]["Addr"]]
    token = swarm_info["JoinTokens"]["Worker"]
    with pytest.raises(DockerError):
        await swarm.swarm.join(join_token=token, remote_addrs=swarm_addr)


@pytest.mark.asyncio
async def test_swarm_init(docker):
    if sys.platform == "win32":
        pytest.skip("swarm commands dont work on Windows")
    default_addr_pool = ["10.0.0.0/8"]
    data_path_port = 4789
    subnet_size = 24
    assert await docker.swarm.init(
        advertise_addr="127.0.0.1:2377",
        data_path_port=data_path_port,
        default_addr_pool=default_addr_pool,
        subnet_size=subnet_size,
        force_new_cluster=True,
    )
    swarm_inspect = await docker.swarm.inspect()
    assert swarm_inspect["DefaultAddrPool"] == default_addr_pool
    assert swarm_inspect["DataPathPort"] == data_path_port
    assert swarm_inspect["SubnetSize"] == subnet_size
    await docker.swarm.leave(force=True)



from __future__ import annotations

import json
from typing import Any, Dict, List, Mapping, Optional, Sequence

from .utils import clean_filters


class DockerNetworks:
    def __init__(self, docker):
        self.docker = docker

    async def list(
        self,
        *,
        filters: Optional[Mapping[str, str | Sequence[str]]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Return a list of networks

        Args:
            filters: a dict with a list of filters

        Available filters:
            dangling=<boolean>
            driver=<driver-name>
            id=<network-id>
            label=<key> or label=<key>=<value> of a network label.
            name=<network-name>
            scope=["swarm"|"global"|"local"]
            type=["custom"|"builtin"]
        """
        params = {} if filters is None else {"filters": clean_filters(filters)}

        data = await self.docker._query_json("networks", params=params)
        return data

    async def create(self, config: Dict[str, Any]) -> DockerNetwork:
        bconfig = json.dumps(config, sort_keys=True).encode("utf-8")
        data = await self.docker._query_json(
            "networks/create", method="POST", data=bconfig
        )
        return DockerNetwork(self.docker, data["Id"])

    async def get(self, net_specs: str) -> DockerNetwork:
        data = await self.docker._query_json(f"networks/{net_specs}", method="GET")
        return DockerNetwork(self.docker, data["Id"])


class DockerNetwork:
    def __init__(self, docker, id_):
        self.docker = docker
        self.id = id_

    async def show(self) -> Dict[str, Any]:
        data = await self.docker._query_json(f"networks/{self.id}")
        return data

    async def delete(self) -> bool:
        async with self.docker._query(f"networks/{self.id}", method="DELETE") as resp:
            return resp.status == 204

    async def connect(self, config: Dict[str, Any]) -> None:
        bconfig = json.dumps(config, sort_keys=True).encode("utf-8")
        await self.docker._query_json(
            f"networks/{self.id}/connect", method="POST", data=bconfig
        )

    async def disconnect(self, config: Dict[str, Any]) -> None:
        bconfig = json.dumps(config, sort_keys=True).encode("utf-8")
        await self.docker._query_json(
            f"networks/{self.id}/disconnect",
            method="POST",
            data=bconfig,
        )



import pytest

from aiodocker.exceptions import DockerError


@pytest.fixture
async def tmp_secret(swarm, random_name):
    secret = await swarm.secrets.create(
        name="secret-" + random_name(), data=random_name()
    )
    yield secret["ID"]
    await swarm.secrets.delete(secret["ID"])


@pytest.mark.asyncio
async def test_secret_list_with_filter(swarm, tmp_secret):
    docker_secret = await swarm.secrets.inspect(secret_id=tmp_secret)
    name = docker_secret["Spec"]["Name"]
    filters = {"name": name}
    filtered_list = await swarm.secrets.list(filters=filters)
    assert len(filtered_list) == 1


@pytest.mark.asyncio
async def test_secret_update(swarm, tmp_secret):
    secret = await swarm.secrets.inspect(secret_id=tmp_secret)
    secret_id = secret["ID"]

    secret = await swarm.secrets.inspect(secret_id)
    current_labels = secret["Spec"]["Labels"]
    assert current_labels == {}
    version = secret["Version"]["Index"]

    # update the secret labels
    await swarm.secrets.update(
        secret_id=secret_id, version=version, labels={"label1": "value1"}
    )

    secret = await swarm.secrets.inspect(secret_id)
    current_labels = secret["Spec"]["Labels"]
    version = secret["Version"]["Index"]
    assert current_labels == {"label1": "value1"}


@pytest.mark.asyncio
async def test_secret_labels(swarm, tmp_secret):
    secret = await swarm.secrets.inspect(secret_id=tmp_secret)
    secret_id1 = secret["ID"]
    secret = await swarm.secrets.inspect(secret_id1)
    version = secret["Version"]["Index"]

    await swarm.secrets.update(
        secret_id=secret_id1, version=version, labels={"label1": "value1"}
    )

    # create a secret with labels
    name = "test_secret2"
    secret = await swarm.secrets.create(
        name=name, data="test secret2", labels={"label2": "value2"}
    )
    secret_id2 = secret["ID"]
    secret = await swarm.secrets.inspect(secret_id2)
    version = secret["Version"]["Index"]
    current_labels = secret["Spec"]["Labels"]
    assert current_labels == {"label2": "value2"}

    # search secret based on labels
    filters = {"label": "label1=value1"}
    filtered_list = await swarm.secrets.list(filters=filters)
    assert len(filtered_list) == 1

    await swarm.secrets.update(
        secret_id=secret_id2, version=version, labels={"label1": "value1"}
    )
    secret = await swarm.secrets.inspect(secret_id2)
    version = secret["Version"]["Index"]
    current_labels = secret["Spec"]["Labels"]
    assert current_labels == {"label1": "value1"}

    filters = {"label": "label1=value1"}
    filtered_list = await swarm.secrets.list(filters=filters)
    assert len(filtered_list) == 2

    await swarm.secrets.delete(secret_id2)


@pytest.mark.asyncio
async def test_secret_update_error(swarm, tmp_secret):
    secret = await swarm.secrets.inspect(secret_id=tmp_secret)
    secret_id = secret["ID"]

    secret = await swarm.secrets.inspect(secret_id)
    version = secret["Version"]["Index"]

    with pytest.raises(DockerError) as error:
        await swarm.secrets.update(
            secret_id=secret_id, version=version, name="new name"
        )
    assert (
        error.value.message == "rpc error: "
        "code = InvalidArgument "
        "desc = only updates to Labels are allowed"
    )


@pytest.mark.asyncio
async def test_secret_create_nodata_error(swarm):
    name = "test_secret-create_nodata_error"
    with pytest.raises(TypeError):
        await swarm.secrets.create(name=name)


@pytest.mark.asyncio
async def test_secret_create_b64_error(swarm):
    name = "test_secret-create_b64_error"
    not_b64 = "I'm not base64 encoded"
    with pytest.raises(DockerError) as error:
        await swarm.secrets.create(name=name, data=not_b64, b64=True)
    assert "illegal base64 data at input byte 1" in error.value.message


@pytest.mark.asyncio
async def test_secret_create_duplicated_error(swarm, tmp_secret):
    secret = await swarm.secrets.inspect(secret_id=tmp_secret)
    name = secret["Spec"]["Name"]
    with pytest.raises(DockerError) as error:
        await swarm.secrets.create(name=name, data="test secret")
    assert (
        error.value.message == "rpc error: "
        "code = AlreadyExists "
        "desc = secret " + name + " already exists"
    )



from __future__ import annotations

import asyncio

import pytest

from aiodocker.docker import Docker


@pytest.mark.asyncio
async def test_events_default_task(docker: Docker) -> None:
    docker.events.subscribe()
    assert docker.events.task is not None
    await docker.close()
    assert docker.events.task is None
    assert docker.events.json_stream is None


@pytest.mark.asyncio
async def test_events_provided_task(docker: Docker) -> None:
    task = asyncio.ensure_future(docker.events.run())
    docker.events.subscribe(create_task=False)
    assert docker.events.task is None
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task
    await docker.close()
    assert docker.events.json_stream is None


@pytest.mark.asyncio
async def test_events_no_task(docker: Docker) -> None:
    assert docker.events.task is None
    await docker.close()
    assert docker.events.json_stream is None



from __future__ import annotations

import pytest

from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError
from aiodocker.networks import DockerNetwork


@pytest.mark.asyncio
async def test_list_networks(docker: Docker) -> None:
    data = await docker.networks.list()
    networks = {net["Name"]: net for net in data}
    assert "none" in networks
    assert networks["none"]["Driver"] == "null"


@pytest.mark.asyncio
async def test_list_networks_with_filter(docker: Docker) -> None:
    network = await docker.networks.create({
        "Name": "test-net-filter",
        "Labels": {"some": "label"},
    })
    try:
        networks = await docker.networks.list(filters={"label": "some=label"})
        assert len(networks) == 1
    finally:
        if network:
            deleted = await network.delete()
            assert deleted is True


@pytest.mark.asyncio
async def test_networks(docker: Docker) -> None:
    network = await docker.networks.create({"Name": "test-net"})
    net_find = await docker.networks.get("test-net")
    assert (await net_find.show())["Name"] == "test-net"
    assert isinstance(network, DockerNetwork)
    data = await network.show()
    assert data["Name"] == "test-net"
    container = None
    try:
        container = await docker.containers.create({"Image": "python"}, name="test-net")
        await network.connect({"Container": "test-net"})
        await network.disconnect({"Container": "test-net"})
    finally:
        if container is not None:
            await container.delete()
        network_delete_result = await network.delete()
        assert network_delete_result is True


@pytest.mark.asyncio
async def test_network_delete_error(docker: Docker) -> None:
    network = await docker.networks.create({"Name": "test-delete-net"})
    assert await network.delete() is True
    with pytest.raises(DockerError):
        await network.delete()



from __future__ import annotations

from typing import (
    Mapping,
    Protocol,
    Sequence,
    TypedDict,
    TypeVar,
    Union,
)

from typing_extensions import TypeAlias


_T_co = TypeVar("_T_co", covariant=True)


class SupportsRead(Protocol[_T_co]):
    def read(self, length: int = ..., /) -> _T_co: ...


# NOTE: Currently these types are used to annotate arguments only.
# When returning values, we need extra type-narrowing for individual fields,
# so it is better to define per-API typed DTOs.
JSONValue: TypeAlias = Union[
    str,
    int,
    float,
    bool,
    None,
    Mapping[str, "JSONValue"],
    Sequence["JSONValue"],
]
JSONObject: TypeAlias = Mapping[str, "JSONValue"]
JSONList: TypeAlias = Sequence["JSONValue"]


class PortInfo(TypedDict):
    HostIp: str
    HostPort: str



from __future__ import annotations

import os
import sys
from io import BytesIO
from typing import AsyncIterator, Callable

import pytest

from aiodocker import utils
from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError


def skip_windows() -> None:
    if sys.platform == "win32":
        # replaced xfail with skip for sake of tests speed
        pytest.skip("image operation fails on Windows")


@pytest.mark.asyncio
async def test_build_from_remote_file(
    docker: Docker,
    random_name: Callable[[], str],
    requires_api_version: Callable[[str, str], None],
) -> None:
    skip_windows()

    requires_api_version(
        "v1.28",
        "TODO: test disabled because it fails on "
        "API version 1.27, this should be fixed",
    )

    remote = (
        "https://raw.githubusercontent.com/aio-libs/"
        "aiodocker/main/tests/docker/Dockerfile"
    )

    tag = f"{random_name()}:1.0"
    await docker.images.build(tag=tag, remote=remote, stream=False)

    image = await docker.images.inspect(tag)
    assert image


@pytest.mark.asyncio
async def test_build_from_remote_tar(
    docker: Docker, random_name: Callable[[], str]
) -> None:
    skip_windows()

    remote = (
        "https://github.com/aio-libs/aiodocker/"
        "raw/main/tests/docker/docker_context.tar"
    )

    tag = f"{random_name()}:1.0"
    await docker.images.build(tag=tag, remote=remote, stream=False)

    image = await docker.images.inspect(tag)
    assert image


@pytest.mark.asyncio
async def test_history(docker: Docker, image_name: str) -> None:
    history = await docker.images.history(name=image_name)
    assert history


@pytest.mark.asyncio
async def test_list_images(docker: Docker, image_name: str) -> None:
    images = await docker.images.list(filter=image_name)
    assert len(images) >= 1


@pytest.mark.asyncio
async def test_tag_image(
    docker: Docker, random_name: Callable[[], str], image_name: str
) -> None:
    repository = random_name()
    await docker.images.tag(name=image_name, repo=repository, tag="1.0")
    await docker.images.tag(name=image_name, repo=repository, tag="2.0")
    image = await docker.images.inspect(image_name)
    assert len([x for x in image["RepoTags"] if x.startswith(repository)]) == 2


@pytest.mark.asyncio
async def test_push_image(docker: Docker, image_name: str) -> None:
    repository = "localhost:5000/image"
    await docker.images.tag(name=image_name, repo=repository)
    await docker.images.push(name=repository)


@pytest.mark.asyncio
async def test_push_image_stream(docker: Docker, image_name: str) -> None:
    repository = "localhost:5000/image"
    await docker.images.tag(name=image_name, repo=repository)
    async for item in docker.images.push(name=repository, stream=True):
        pass


@pytest.mark.asyncio
async def test_delete_image(docker: Docker, image_name: str) -> None:
    repository = "localhost:5000/image"
    await docker.images.tag(name=image_name, repo=repository)
    assert await docker.images.inspect(repository)
    await docker.images.delete(name=repository)


@pytest.mark.asyncio
async def test_not_existing_image(
    docker: Docker, random_name: Callable[[], str]
) -> None:
    name = f"{random_name()}:latest"
    with pytest.raises(DockerError) as excinfo:
        await docker.images.inspect(name=name)
    assert excinfo.value.status == 404


@pytest.mark.asyncio
async def test_pull_image(docker: Docker, image_name: str) -> None:
    image = await docker.images.inspect(name=image_name)
    assert image

    with pytest.warns(DeprecationWarning):
        image = await docker.images.get(name=image_name)
        assert "Architecture" in image


@pytest.mark.asyncio
async def test_pull_image_stream(docker: Docker, image_name: str) -> None:
    image = await docker.images.inspect(name=image_name)
    assert image

    async for item in docker.images.pull(image_name, stream=True):
        pass


@pytest.mark.asyncio
async def test_build_from_tar(
    docker: Docker, random_name: Callable[[], str], image_name: str
) -> None:
    name = f"{random_name()}:latest"
    dockerfile = f"""
    # Shared Volume
    FROM {image_name}
    """
    f = BytesIO(dockerfile.encode("utf-8"))
    tar_obj = utils.mktar_from_dockerfile(f)
    await docker.images.build(fileobj=tar_obj, encoding="gzip", tag=name)
    tar_obj.close()
    image = await docker.images.inspect(name=name)
    assert image


@pytest.mark.asyncio
async def test_build_from_tar_stream(
    docker: Docker, random_name: Callable[[], str], image_name: str
) -> None:
    name = f"{random_name()}:latest"
    dockerfile = f"""
    # Shared Volume
    FROM {image_name}
    """
    f = BytesIO(dockerfile.encode("utf-8"))
    tar_obj = utils.mktar_from_dockerfile(f)
    async for item in docker.images.build(
        fileobj=tar_obj, encoding="gzip", tag=name, stream=True
    ):
        pass
    tar_obj.close()
    image = await docker.images.inspect(name=name)
    assert image


@pytest.mark.asyncio
async def test_export_image(docker: Docker, image_name: str) -> None:
    name = image_name
    async with docker.images.export_image(name=name) as exported_image:
        assert exported_image
        async for chunk in exported_image.iter_chunks():
            pass


@pytest.mark.asyncio
async def test_import_image(docker: Docker) -> None:
    skip_windows()

    async def file_sender(file_name: str) -> AsyncIterator[bytes]:
        with open(file_name, "rb") as f:
            chunk = f.read(2**16)
            while chunk:
                yield chunk
                chunk = f.read(2**16)

    dir = os.path.dirname(__file__)
    hello_world = os.path.join(dir, "docker/google-containers-pause.tar")
    # FIXME: improve annotation for chunked data generator
    response = await docker.images.import_image(data=file_sender(hello_world))  # type: ignore
    for item in response:
        assert "error" not in item

    repository = "gcr.io/google-containers/pause"

    for tag in ["1.0", "go", "latest", "test", "test2"]:
        name = f"{repository}:{tag}"
        image = await docker.images.inspect(name=name)
        assert image


@pytest.mark.asyncio
async def test_pups_image_auth(docker: Docker, image_name: str) -> None:
    skip_windows()

    name = image_name
    await docker.images.pull(from_image=name)
    repository = "localhost:5001/image:latest"
    image, tag = repository.rsplit(":", 1)
    registry_addr, image_name = image.split("/", 1)
    await docker.images.tag(name=name, repo=image, tag=tag)

    auth_config = {"username": "testuser", "password": "testpassword"}

    await docker.images.push(name=repository, tag=tag, auth=auth_config)

    await docker.images.delete(name=repository)
    await docker.images.pull(repository, auth={"auth": "dGVzdHVzZXI6dGVzdHBhc3N3b3Jk"})

    await docker.images.inspect(repository)
    await docker.images.delete(name=repository)

    # Now compose_auth_header automatically parse and rebuild
    # the encoded value if required.
    await docker.pull(repository, auth="dGVzdHVzZXI6dGVzdHBhc3N3b3Jk")
    with pytest.raises(ValueError):
        # The repository arg must include the registry address.
        await docker.pull("image:latest", auth={"auth": "dGVzdHVzZXI6dGVzdHBhc3N3b3Jk"})
    await docker.pull(repository, auth={"auth": "dGVzdHVzZXI6dGVzdHBhc3N3b3Jk"})
    await docker.images.inspect(repository)


@pytest.mark.asyncio
async def test_build_image_invalid_platform(docker: Docker, image_name: str) -> None:
    dockerfile = f"""
        FROM {image_name}
        """
    f = BytesIO(dockerfile.encode("utf-8"))
    tar_obj = utils.mktar_from_dockerfile(f)
    with pytest.raises(DockerError) as excinfo:
        async for item in docker.images.build(
            fileobj=tar_obj, encoding="gzip", stream=True, platform="foo"
        ):
            pass
    tar_obj.close()
    assert excinfo.value.status == 400
    assert (
        "unknown operating system or architecture: invalid argument"
        in excinfo.exconly()
    )


@pytest.mark.asyncio
async def test_pull_image_invalid_platform(docker: Docker, image_name: str) -> None:
    with pytest.raises(DockerError) as excinfo:
        await docker.images.pull("hello-world", platform="foo")

    assert excinfo.value.status == 400
    assert (
        "unknown operating system or architecture: invalid argument"
        in excinfo.exconly()
    )



from __future__ import annotations

import json

from .utils import clean_filters


class DockerVolumes:
    def __init__(self, docker):
        self.docker = docker

    async def list(self, *, filters=None):
        """
        Return a list of volumes

        Args:
            filters: a dict with a list of filters

        Available filters:
            dangling=<boolean>
            driver=<volume-driver-name>
            label=<key> or label=<key>:<value>
            name=<volume-name>
        """
        params = {} if filters is None else {"filters": clean_filters(filters)}

        data = await self.docker._query_json("volumes", params=params)
        return data

    async def get(self, id):
        data = await self.docker._query_json(f"volumes/{id}", method="GET")
        return DockerVolume(self.docker, data["Name"])

    async def create(self, config):
        config = json.dumps(config, sort_keys=True).encode("utf-8")
        data = await self.docker._query_json(
            "volumes/create", method="POST", data=config
        )
        return DockerVolume(self.docker, data["Name"])


class DockerVolume:
    def __init__(self, docker, name):
        self.docker = docker
        self.name = name

    async def show(self):
        data = await self.docker._query_json(f"volumes/{self.name}")
        return data

    async def delete(self, force=False):
        async with self.docker._query(
            f"volumes/{self.name}", method="DELETE", params=dict(force=force)
        ):
            pass



import pytest

from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError


@pytest.mark.asyncio
async def test_create_search_get_delete(docker: Docker) -> None:
    name = "aiodocker-test-volume-two"
    await docker.volumes.create({
        "Name": name,
        "Labels": {"some": "label"},
        "Driver": "local",
    })
    volumes_response = await docker.volumes.list(filters={"label": "some=label"})
    volumes = volumes_response["Volumes"]
    assert len(volumes) == 1
    volume_data = volumes[0]
    volume = await docker.volumes.get(volume_data["Name"])
    await volume.delete()
    with pytest.raises(DockerError):
        await docker.volumes.get(name)


@pytest.mark.asyncio
@pytest.mark.parametrize("force_delete", [True, False])
async def test_create_show_delete_volume(docker: Docker, force_delete: bool) -> None:
    name = "aiodocker-test-volume"
    volume = await docker.volumes.create({
        "Name": name,
        "Labels": {},
        "Driver": "local",
    })
    assert volume
    data = await volume.show()
    assert data
    await volume.delete(force_delete)
    with pytest.raises(DockerError):
        await docker.volumes.get(name)



#!/usr/bin/env python3

import asyncio

from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError


async def demo(docker):
    try:
        await docker.images.inspect("alpine:latest")
    except DockerError as e:
        if e.status == 404:
            await docker.pull("alpine:latest")
        else:
            print("Error retrieving alpine:latest image.")
            return

    subscriber = docker.events.subscribe()

    config = {
        "Cmd": ["tail", "-f", "/var/log/dmesg"],
        "Image": "alpine:latest",
        "AttachStdin": False,
        "AttachStdout": True,
        "AttachStderr": True,
        "Tty": False,
        "OpenStdin": False,
        "StdinOnce": False,
    }
    container = await docker.containers.create_or_replace(config=config, name="testing")
    await container.start()
    print(f"=> created and started container {container._id[:12]}")

    while True:
        event = await subscriber.get()
        if event is None:
            break

        for key, value in event.items():
            print(key, ":", value)

        # Demonstrate simple event-driven container mgmt.
        if event["Actor"]["ID"] == container._id:
            if event["Action"] == "start":
                await container.stop()
                print(f"=> killed {container._id[:12]}")
            elif event["Action"] == "stop":
                await container.delete(force=True)
                print(f"=> deleted {container._id[:12]}")
            elif event["Action"] == "destroy":
                print("=> done with this container!")
                break


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    docker = Docker()
    try:
        # do our stuffs.
        loop.run_until_complete(demo(docker))
    finally:
        loop.run_until_complete(docker.close())
        loop.close()



from __future__ import annotations

import json
import shlex
import tarfile
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

from aiohttp import ClientWebSocketResponse
from multidict import MultiDict
from yarl import URL

from aiodocker.types import JSONObject

from .exceptions import DockerContainerError, DockerError
from .execs import Exec
from .jsonstream import json_stream_list, json_stream_stream
from .logs import DockerLog
from .multiplexed import multiplexed_result_list, multiplexed_result_stream
from .stream import Stream
from .types import PortInfo
from .utils import identical, parse_result


if TYPE_CHECKING:
    from .docker import Docker


class DockerContainers:
    def __init__(self, docker: Docker) -> None:
        self.docker = docker

    async def list(self, **kwargs) -> List[DockerContainer]:
        data = await self.docker._query_json(
            "containers/json", method="GET", params=kwargs
        )
        return [DockerContainer(self.docker, **x) for x in data]

    async def create_or_replace(
        self,
        name: str,
        config: JSONObject,
    ) -> DockerContainer:
        container = None

        try:
            container = await self.get(name)
            if not identical(config, container._container):
                running = container._container.get("State", {}).get("Running", False)
                if running:
                    await container.stop()
                await container.delete()
                container = None
        except DockerError:
            pass

        if container is None:
            container = await self.create(config, name=name)

        return container

    async def create(
        self,
        config: JSONObject,
        *,
        name: Optional[str] = None,
    ) -> DockerContainer:
        url = "containers/create"
        encoded_config = json.dumps(config, sort_keys=True).encode("utf-8")
        kwargs = {}
        if name:
            kwargs["name"] = name
        data = await self.docker._query_json(
            url, method="POST", data=encoded_config, params=kwargs
        )
        return DockerContainer(self.docker, id=data["Id"])

    async def run(
        self,
        config: JSONObject,
        *,
        auth: Optional[Union[Mapping, str, bytes]] = None,
        name: Optional[str] = None,
    ) -> DockerContainer:
        """
        Create and start a container.

        If container.start() will raise an error the exception will contain
        a `container_id` attribute with the id of the container.

        Use `auth` for specifying credentials for pulling absent image from
        a private registry.
        """
        try:
            container = await self.create(config, name=name)
        except DockerError as err:
            # image not found, try pulling it
            if err.status == 404 and "Image" in config:
                await self.docker.pull(str(config["Image"]), auth=auth)
                container = await self.create(config, name=name)
            else:
                raise err

        try:
            await container.start()
        except DockerError as err:
            raise DockerContainerError(
                err.status, {"message": err.message}, container["id"]
            )

        return container

    async def get(self, container_id: str, **kwargs) -> DockerContainer:
        data = await self.docker._query_json(
            f"containers/{container_id}/json",
            method="GET",
            params=kwargs,
        )
        return DockerContainer(self.docker, **data)

    def container(self, container_id: str, **kwargs) -> DockerContainer:
        data = {"id": container_id}
        data.update(kwargs)
        return DockerContainer(self.docker, **data)

    def exec(self, exec_id: str) -> Exec:
        """Return Exec instance for already created exec object."""
        return Exec(self.docker, exec_id)


class DockerContainer:
    _container: Dict[str, Any]

    def __init__(self, docker: Docker, **kwargs) -> None:
        self.docker = docker
        self._container = kwargs
        self._id = self._container.get(
            "id", self._container.get("ID", self._container.get("Id"))
        )
        self.logs = DockerLog(docker, self)

    @property
    def id(self) -> str:
        return self._id

    @overload
    async def log(
        self,
        *,
        stdout: bool = False,
        stderr: bool = False,
        follow: Literal[False] = False,
        **kwargs,
    ) -> List[str]: ...

    @overload
    def log(
        self,
        *,
        stdout: bool = False,
        stderr: bool = False,
        follow: Literal[True],
        **kwargs,
    ) -> AsyncIterator[str]: ...

    def log(
        self,
        *,
        stdout: bool = False,
        stderr: bool = False,
        follow: bool = False,
        **kwargs,
    ) -> Any:
        if stdout is False and stderr is False:
            raise TypeError("Need one of stdout or stderr")

        params = {"stdout": stdout, "stderr": stderr, "follow": follow}
        params.update(kwargs)
        cm = self.docker._query(
            f"containers/{self._id}/logs", method="GET", params=params
        )
        if follow:
            return self._logs_stream(cm)
        else:
            return self._logs_list(cm)

    async def _logs_stream(self, cm):
        try:
            inspect_info = await self.show()
        except DockerError:
            cm.cancel()
            raise
        is_tty = inspect_info["Config"]["Tty"]

        async with cm as response:
            async for item in multiplexed_result_stream(response, is_tty=is_tty):
                yield item

    async def _logs_list(self, cm):
        try:
            inspect_info = await self.show()
        except DockerError:
            raise
        is_tty = inspect_info["Config"]["Tty"]

        async with cm as response:
            return await multiplexed_result_list(response, is_tty=is_tty)

    async def get_archive(self, path: str) -> tarfile.TarFile:
        async with self.docker._query(
            f"containers/{self._id}/archive",
            method="GET",
            params={"path": path},
        ) as response:
            data = await parse_result(response)
            return data

    async def put_archive(self, path, data):
        async with self.docker._query(
            f"containers/{self._id}/archive",
            method="PUT",
            data=data,
            headers={"content-type": "application/json"},
            params={"path": path},
        ) as response:
            data = await parse_result(response)
            return data

    async def show(self, **kwargs) -> Dict[str, Any]:
        data = await self.docker._query_json(
            f"containers/{self._id}/json", method="GET", params=kwargs
        )
        self._container = data
        return data

    async def stop(self, **kwargs) -> None:
        async with self.docker._query(
            f"containers/{self._id}/stop", method="POST", params=kwargs
        ):
            pass

    async def start(self, **kwargs) -> None:
        async with self.docker._query(
            f"containers/{self._id}/start",
            method="POST",
            headers={"content-type": "application/json"},
            params=kwargs,
        ):
            pass

    async def restart(self, timeout=None) -> None:
        params = {}
        if timeout is not None:
            params["t"] = timeout
        async with self.docker._query(
            f"containers/{self._id}/restart",
            method="POST",
            params=params,
        ):
            pass

    async def kill(self, **kwargs) -> None:
        async with self.docker._query(
            f"containers/{self._id}/kill", method="POST", params=kwargs
        ):
            pass

    async def wait(self, *, timeout=None, **kwargs) -> Dict[str, Any]:
        data = await self.docker._query_json(
            f"containers/{self._id}/wait",
            method="POST",
            params=kwargs,
            timeout=timeout,
        )
        return data

    async def delete(self, **kwargs) -> None:
        async with self.docker._query(
            f"containers/{self._id}", method="DELETE", params=kwargs
        ):
            pass

    async def rename(self, newname) -> None:
        async with self.docker._query(
            f"containers/{self._id}/rename",
            method="POST",
            headers={"content-type": "application/json"},
            params={"name": newname},
        ):
            pass

    async def websocket(self, **params) -> ClientWebSocketResponse:
        if not params:
            params = {"stdin": True, "stdout": True, "stderr": True, "stream": True}
        path = f"containers/{self._id}/attach/ws"
        ws = await self.docker._websocket(path, **params)
        return ws

    def attach(
        self,
        *,
        stdout: bool = False,
        stderr: bool = False,
        stdin: bool = False,
        detach_keys: Optional[str] = None,
        logs: bool = False,
    ) -> Stream:
        async def setup() -> Tuple[URL, Optional[bytes], bool]:
            params: MultiDict[Union[str, int]] = MultiDict()
            if detach_keys:
                params.add("detachKeys", detach_keys)
            else:
                params.add("detachKeys", "")
            params.add("logs", int(logs))
            params.add("stdin", int(stdin))
            params.add("stdout", int(stdout))
            params.add("stderr", int(stderr))
            params.add("stream", 1)
            inspect_info = await self.show()
            return (
                URL(f"containers/{self._id}/attach").with_query(params),
                None,
                inspect_info["Config"]["Tty"],
            )

        return Stream(self.docker, setup, None)

    async def port(self, private_port: int | str) -> List[PortInfo] | None:
        if "NetworkSettings" not in self._container:
            await self.show()

        private_port = str(private_port)
        h_ports = None

        # Port settings is None when the container is running with
        # network_mode=host.
        port_settings = self._container.get("NetworkSettings", {}).get("Ports")
        if port_settings is None:
            return None

        if "/" in private_port:
            return port_settings.get(private_port)

        h_ports = port_settings.get(private_port + "/tcp")
        if h_ports is None:
            h_ports = port_settings.get(private_port + "/udp")

        return h_ports

    @overload
    def stats(
        self,
        *,
        stream: Literal[True] = True,
    ) -> AsyncIterator[Dict[str, Any]]: ...

    @overload
    async def stats(
        self,
        *,
        stream: Literal[False],
    ) -> List[Dict[str, Any]]: ...

    def stats(
        self,
        *,
        stream: bool = True,
    ) -> Any:
        cm = self.docker._query(
            f"containers/{self._id}/stats",
            params={"stream": "1" if stream else "0"},
        )
        if stream:
            return self._stats_stream(cm)
        else:
            return self._stats_list(cm)

    async def _stats_stream(self, cm):
        async with cm as response:
            async for item in json_stream_stream(response):
                yield item

    async def _stats_list(self, cm):
        async with cm as response:
            return await json_stream_list(response)

    async def exec(
        self,
        cmd: Union[str, Sequence[str]],
        stdout: bool = True,
        stderr: bool = True,
        stdin: bool = False,
        tty: bool = False,
        privileged: bool = False,
        user: str = "",  # root by default
        environment: Optional[Union[Mapping[str, str], Sequence[str]]] = None,
        workdir: Optional[str] = None,
        detach_keys: Optional[str] = None,
    ) -> Exec:
        if isinstance(cmd, str):
            cmd = shlex.split(cmd)
        if environment is None:
            pass
        elif isinstance(environment, Mapping):
            environment = [f"{key}={value}" for key, value in environment.items()]
        else:
            environment = list(environment)
        data = {
            "Container": self._id,
            "Privileged": privileged,
            "Tty": tty,
            "AttachStdin": stdin,
            "AttachStdout": stdout,
            "AttachStderr": stderr,
            "Cmd": cmd,
            "Env": environment,
        }

        if workdir is not None:
            data["WorkingDir"] = workdir
        else:
            data["WorkingDir"] = ""

        if detach_keys:
            data["detachKeys"] = detach_keys
        else:
            data["detachKeys"] = ""

        if user:
            data["User"] = user
        else:
            data["User"] = ""

        data = await self.docker._query_json(
            f"containers/{self._id}/exec", method="POST", data=data
        )
        return Exec(self.docker, data["Id"], tty=tty)

    async def resize(self, *, h: int, w: int) -> None:
        url = URL(f"containers/{self._id}/resize").with_query(h=h, w=w)
        await self.docker._query_json(url, method="POST")

    async def commit(
        self,
        *,
        repository: Optional[str] = None,
        tag: Optional[str] = None,
        message: Optional[str] = None,
        author: Optional[str] = None,
        changes: Optional[Union[str, Sequence[str]]] = None,
        config: Optional[Dict[str, Any]] = None,
        pause: bool = True,
    ) -> Dict[str, Any]:
        """
        Commit a container to an image. Similar to the ``docker commit``
        command.
        """
        params = {"container": self._id, "pause": pause}
        if repository is not None:
            params["repo"] = repository
        if tag is not None:
            params["tag"] = tag
        if message is not None:
            params["comment"] = message
        if author is not None:
            params["author"] = author
        if changes is not None:
            if not isinstance(changes, str):
                changes = "\n".join(changes)
            params["changes"] = changes

        return await self.docker._query_json(
            "commit", method="POST", params=params, data=config
        )

    async def pause(self) -> None:
        async with self.docker._query(f"containers/{self._id}/pause", method="POST"):
            pass

    async def unpause(self) -> None:
        async with self.docker._query(f"containers/{self._id}/unpause", method="POST"):
            pass

    def __getitem__(self, key: str) -> Any:
        return self._container[key]

    def __hasitem__(self, key: str) -> bool:
        return key in self._container
